/*
    Copyright 2007 Robert Knight <robertknight@gmail.com>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public License
    along with this library; see the file COPYING.LIB.  If not, write to
    the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
    Boston, MA 02110-1301, USA.
*/

// Own
#include "../core/recentlyusedapplicationsmodel.h"

// Qt
#include <QtCore/QFileInfo>

// KDE
#include <KDesktopFile>
#include <KDirWatch>
#include <KIcon>
#include <KLocalizedString>
#include <KRecentDocument>
#include <KUrl>
#include <KDebug>

// Local
#include "../core/models.h"
#include "../core/recentapplications.h"
#include "../recentappadaptor.h"

using namespace Kickoff;

class RecentlyUsedApplicationsModel::Private
{
public:
    Private(RecentlyUsedApplicationsModel *parent)
            : q(parent){
    }
    void removeExistingItem(const QString& path) {
        if (!itemsByPath.contains(path)) {
            return;
        }

        QStandardItem *existingItem = itemsByPath[path];
        //kDebug() << "Removing existing item" << existingItem;
        Q_ASSERT(existingItem->parent());
        existingItem->parent()->removeRow(existingItem->row());
        itemsByPath.remove(path);
    }

    void addRecentApplication(KService::Ptr service, bool append) {
        // remove existing item if any
        removeExistingItem(service->entryPath());

        QStandardItem *appItem = StandardItemFactory::createItemForService(service);
        itemsByPath.insert(service->entryPath(), appItem);

        if (append) {
            recentAppItem->appendRow(appItem);
        } else {
            recentAppItem->insertRow(0, appItem);
        }
    }

    void loadRecentApplications() {
        recentAppItem = new QStandardItem(i18n("Applications"));
        QList<KService::Ptr> services = RecentApplications::self()->recentApplications();
        foreach(const KService::Ptr& service, services) {
            addRecentApplication(service, true);
        }
        q->appendRow(recentAppItem);
    }

    RecentlyUsedApplicationsModel * const q;
    QStandardItem *recentAppItem;

    QHash<QString, QStandardItem*> itemsByPath;
};

RecentlyUsedApplicationsModel::RecentlyUsedApplicationsModel(QObject *parent)
        : KickoffModel(parent)
        , d(new Private(this))
{
    QDBusConnection dbus = QDBusConnection::sessionBus();
    (void)new RecentappAdaptor(this);
    QDBusConnection::sessionBus().registerObject("/kickoff/RecentAppDoc", this);
    dbus.connect(QString(), "/kickoff/RecentAppDoc", "org.kde.plasma", "clearRecentApplications", this, SLOT(clearRecentApplications()));

    d->loadRecentApplications();

    // listen for changes to the list of recent applications
    connect(RecentApplications::self(), SIGNAL(applicationAdded(KService::Ptr, int)),
            this, SLOT(recentApplicationAdded(KService::Ptr, int)));
    connect(RecentApplications::self(), SIGNAL(applicationRemoved(KService::Ptr)),
            this, SLOT(recentApplicationRemoved(KService::Ptr)));
    connect(RecentApplications::self(), SIGNAL(cleared()),
            this, SLOT(recentApplicationsCleared()));
}
RecentlyUsedApplicationsModel::~RecentlyUsedApplicationsModel()
{
    delete d;
}

void RecentlyUsedApplicationsModel::recentApplicationAdded(KService::Ptr service, int)
{
    if (service) {
        d->addRecentApplication(service, false);
    }
}

void RecentlyUsedApplicationsModel::recentApplicationRemoved(KService::Ptr service)
{
    if (service) {
        d->removeExistingItem(service->entryPath());
    }
}

void RecentlyUsedApplicationsModel::recentApplicationsCleared()
{
    QSet<QStandardItem*> appItems;
    const int rows = d->recentAppItem->rowCount();
    for (int i = 0;i < rows;i++) {
        appItems << d->recentAppItem->child(i);
    }
    QMutableHashIterator<QString, QStandardItem*> iter(d->itemsByPath);
    while (iter.hasNext()) {
        iter.next();
        if (appItems.contains(iter.value())) {
            iter.remove();
        }
    }

    d->recentAppItem->removeRows(0, d->recentAppItem->rowCount());
}
void RecentlyUsedApplicationsModel::clearRecentApplications()
{
    RecentApplications::self()->clear();
}


#include "recentlyusedapplicationsmodel.moc"

